#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <i86.h>
#include <time.h>
#include "tiles.h"
#include "map.h"

/* VGA sequence controller */
#define SC_INDEX      0x03c4
#define SC_DATA       0x03c5


// VGA Graphics Registers
// http://www.osdever.net/FreeVGA/vga/graphreg.htm
#define GAR_INDEX	  0x03CE
#define GAR_DATA	  0x03CF

#define ACR_INDEX	  0x03C0
#define ACR_DATA	  0x03C1


/* VGA CRT controller */
#define CRTC_INDEX    0x03d4
#define CRTC_DATA     0x03d5

#define MEMORY_MODE   0x04
#define UNDERLINE_LOC 0x14
#define MODE_CONTROL  0x17

/* CRT controller registers */
#define HIGH_ADDRESS 0x0C
#define LOW_ADDRESS  0x0D

#define MAXSPRITES 16


//  BUGS:
// 	- Sprites leave garbage behind at some moments during scrolling
//  - Keyboard can't move in certain diagonals and press space simultaneously


#define SPRITE_SPEED 2
#define ROBOT_SPEED 2


#define TILE_FLOPPYDISK 162
#define TILE_FLOOR 122
#define TILE_EXIT 81


struct sprite_struct
{
	signed char type;
	int x;
	int y;
	int lastx;
	int lasty;
	int xvel;
	int yvel;
	int lastxvel;
	int lastyvel;
	char dir;
	unsigned char tile;
	unsigned char animCounter;
	unsigned char background[256];	
} sprites[MAXSPRITES];

struct sprite_struct *player;

char markerTimer = 0;

char gameRunning = 1;
char goNextLevel = 0;
char goToStart = 0;

char keyboardUp = 0;
char keyboardDown = 0;
char keyboardLeft = 0;
char keyboardRight = 0;
char keyboardA = 0;
char keyboardD = 0;
char keyboardW = 0;
char keyboardS = 0;
char keyboardShiftL = 0;
char keyboardShiftR = 0;
char keyboardLast = 0;
char keyboardAnyKey = 0;

char currentLevel = 0;

char logicEnabled;
char noClip = 0;
char floppiesCollected = 0;
char placeMarker = 0;
char skipRender = 0;

int tileUpdates[100];
unsigned char tileUpdatePosition;

unsigned char frameCounter = 0;

//__segment videoSegment = 0xa000; //0xb800;
//char __based(void) *videoOffset = 0;
//char far *screenptr = (char far*) MK_FP(0xA000,0x0000); 

unsigned char far *videoMem;
unsigned char originalPalette[256 * 3];
unsigned char currentPalette[256 * 3];

unsigned int visiblePage;
unsigned int nonvisiblePage;
unsigned int videoStorage;
unsigned int tempPage;
unsigned int high_addr;
unsigned int low_addr;
unsigned int page_addr;
unsigned int scanlineWidthInBytes;


void (interrupt far *oldkbhandler)(void);
char lastChar = 0;

int screen_xpixeloffset;
int screen_ypixeloffset;

union REGPACK regs;



int mouseX = 0;
int mouseY = 0;
int mouseButtons = 0;
int mouseLastX = 0;
int mouseLastY = 0;
int mouseLastButtons = 0;

int manualScrollTimer = 0;

int map_startX = 0;
int map_startY = 0;

int maze_startX;
int maze_startY;
int maze_endX;
int maze_endY;

unsigned int *currentSound = NULL;

char fadeout = 0;

char asciiTiles[256];

int soundHit[] = { 50, 100, 150, 200, 0 };
int soundFloppy[] = { 500, 550, 600, 650, 700, 0 };
int soundSteps[] = { 400, -1, 200, -1, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
					 400, -1, 200, -1, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
					 400, -1, 200, -1, 0 };
int soundAllFloppies[] = {  800, -1, -1, -1, -1, -1, -1, -1, -1,
							900, -1, -1, -1, -1, -1, -1, -1, -1,
							1000, -1, -1, -1, -1, -1, -1, -1, -1, 
							0 };
int soundMarker[] = {400, 100, -1, 0};

char copyright[] = "COPYRIGHT (C) 2014 JEFFREY BOWMAN";

unsigned char grid[] = {
				1, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0,
				0, 1, 1, 1, 0, 1, 0 ,1 ,0 ,1, 1, 0, 0, 1, 1, 1, 0 ,1 ,1 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 1, 0, 0, 0, 1, 0 ,1 ,0 ,1, 0, 1, 0, 1, 0, 0, 0 ,1 ,0 ,1, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 1, 1, 0, 0, 0, 1 ,0 ,0 ,1, 1, 0, 0, 1, 1, 0, 0 ,1 ,1 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 1, 0, 0, 0, 0, 1 ,0 ,0 ,1, 0, 1, 0, 1, 0, 0, 0 ,1 ,0 ,1, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 1, 0, 0, 0, 0, 1 ,0 ,0 ,1, 1, 0, 0, 1, 1, 1, 0 ,1 ,0 ,1, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 1, 2, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 17, 18, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				17, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 
				0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0, 0, 0, 0, 0, 0, 0, 0 ,0 ,0 ,0				
				};


				


void soundOn(int freq)
{
 	unsigned long Div;
 	char c;
 
    // Program PIT
 	Div = 1193180 / freq;
 	outp(0x43, 0xb6);
 	outp(0x42, Div & 0xFF );
 	outp(0x42, (Div >> 8) & 0xFF);
 
	// Connect speaker to PIT output
 	c = inp(0x61);
  	if (c != (c | 3)) outp(0x61, c | 3);
}
 
void soundOff()
{
 	// Disconnect speaker from PIT
	char c = inp(0x61) & 0xFC; 
 	outp(0x61, c);
}				
				
				



void blitTile8(int x, int y, int tile)
{
	int cx;
	int cy;
	//unsigned char pixel;
	int plane;
	unsigned int memAddress;
	unsigned int tileAddress;
	unsigned int baseTileAddress;
	unsigned int baseMemAddress;
	unsigned int tileBase = (tile << 6);
	//unsigned int xoffset;
	
	baseMemAddress = nonvisiblePage + ((y) << 6) + ((y) << 4);// + (x >> 2);
	baseTileAddress = tileBase;
	
	for (cx = 0; cx < 4; cx++)
	{ 
		plane = ((x + cx) & 3);
		/* select the map mask register */
		outp(SC_INDEX, 0x02);
		/* write 2^plane */
		outp(SC_DATA, 1 << plane);
		
		memAddress = baseMemAddress + ((x + cx) >> 2);
		//memAddress = baseMemAddress;
		tileAddress = baseTileAddress;
		 
		for (cy = 0; cy < 8; cy++)		
		{				
			//*(unsigned short*)(videoMem + memAddress) = tiles[tileAddress];// | (tiles[tileAddress + 4u] << 8) ;	
			videoMem[memAddress] = tiles[tileAddress];				
			videoMem[memAddress + 1] = tiles[tileAddress + 4];
			
			tileAddress += 8u;
			memAddress += scanlineWidthInBytes;
		}
		
		baseTileAddress++;
  
	}		
}



void blitTile16(int x, int y, int tile)
{
	int cx;
	int cy;
	unsigned char pixel;
	int plane;
	unsigned int memAddress;
	unsigned int tileAddress;
	unsigned int baseTileAddress;
	unsigned int baseMemAddress;
	unsigned int tileBase = (tile << 8);
	//unsigned int xoffset;
	
	baseMemAddress = nonvisiblePage + (y * scanlineWidthInBytes);
	baseTileAddress = tileBase;
	
	for (cx = 0; cx < 4; cx++)
	{ 
		plane = ((x + cx) & 3);
		/* select the map mask register */
		outp(SC_INDEX, 0x02);
		/* write 2^plane */
		outp(SC_DATA, 1 << plane);
		
		memAddress = baseMemAddress + ((x + cx) >> 2);
		//memAddress = baseMemAddress;
		tileAddress = baseTileAddress;
		 
		for (cy = 0; cy < 16; cy++)		
		{				
			//*(unsigned short*)(videoMem + memAddress) = tiles[tileAddress];// | (tiles[tileAddress + 4u] << 8) ;	
			pixel = tiles[tileAddress];
			if (pixel != 0xFF) videoMem[memAddress] = pixel;				
			pixel = tiles[tileAddress + 4];
			if (pixel != 0xFF) videoMem[memAddress + 1] = pixel;
			pixel = tiles[tileAddress + 8];
			if (pixel != 0xFF) videoMem[memAddress + 2] = pixel;
			pixel = tiles[tileAddress + 12];
			if (pixel != 0xFF) videoMem[memAddress + 3] = pixel;
			
			tileAddress += 16u;
			memAddress += scanlineWidthInBytes;
		}
		
		baseTileAddress++;	
  
	}		
	

}


void blitTileFrom16(int x, int y, unsigned char *pixels)
{
	int cx;
	int cy;
	unsigned char pixel;
	int plane;
	unsigned int memAddress;
	unsigned int tileAddress;
	unsigned int baseTileAddress;
	unsigned int baseMemAddress;
	unsigned int tileBase = (unsigned int)pixels;
	//unsigned int xoffset;
	
	baseMemAddress = nonvisiblePage + (y * scanlineWidthInBytes);
	baseTileAddress = tileBase;
	
	for (cx = 0; cx < 4; cx++)
	{ 
		plane = ((x + cx) & 3);
		/* select the map mask register */
		//outp(SC_INDEX, 0x02);
		/* write 2^plane */
		//outp(SC_DATA, 1 << plane);
		outpw(SC_INDEX, ((1 << plane) << 8) | 0x02);
		
		memAddress = baseMemAddress + ((x + cx) >> 2);
		//memAddress = baseMemAddress;
		tileAddress = baseTileAddress;
		 
		for (cy = 0; cy < 16; cy++)		
		{				
			//*(unsigned short*)(videoMem + memAddress) = tiles[tileAddress];// | (tiles[tileAddress + 4u] << 8) ;	
			pixel = tiles[tileAddress];
			if (pixel != 0xFF) videoMem[memAddress] = pixel;				
			pixel = tiles[tileAddress + 4];
			if (pixel != 0xFF) videoMem[memAddress + 1] = pixel;
			pixel = tiles[tileAddress + 8];
			if (pixel != 0xFF) videoMem[memAddress + 2] = pixel;
			pixel = tiles[tileAddress + 12];
			if (pixel != 0xFF) videoMem[memAddress + 3] = pixel;
			
			tileAddress += 16u;
			memAddress += scanlineWidthInBytes;
		}
		
		baseTileAddress++;	
  
	}		
	

}

void blitTile16NoMask(int x, int y, int tile)
{
	int cx;
	int cy;
	//unsigned char pixel;
	int plane;
	unsigned int memAddress;
	unsigned int tileAddress;
	unsigned int baseTileAddress;
	unsigned int baseMemAddress;
	unsigned int tileBase = (tile << 8);
	//unsigned int xoffset;
	
	baseMemAddress = nonvisiblePage + (y * scanlineWidthInBytes);
	baseTileAddress = tileBase;
	
	for (cx = 0; cx < 4; cx++)
	{ 
		plane = ((x + cx) & 3);
		/* select the map mask register */
		outp(SC_INDEX, 0x02);
		/* write 2^plane */
		outp(SC_DATA, 1 << plane);
		
		memAddress = baseMemAddress + ((x + cx) >> 2);
		//memAddress = baseMemAddress;
		tileAddress = baseTileAddress;
		 
		for (cy = 0; cy < 16; cy++)		
		{				
			videoMem[memAddress] = tiles[tileAddress];			
			videoMem[memAddress + 1] = tiles[tileAddress + 4];			
			videoMem[memAddress + 2] = tiles[tileAddress + 8];			
			videoMem[memAddress + 3] = tiles[tileAddress + 12];
			
			tileAddress += 16u;
			memAddress += scanlineWidthInBytes;
		}
		
		baseTileAddress++;	
  
	}		
	

}


void saveTile16(int x, int y, int tile)
{
	int cx;
	int cy;
	//unsigned char pixel;
	int plane;
	unsigned int memAddress;
	unsigned int tileAddress;
	unsigned int baseTileAddress;
	unsigned int baseMemAddress;
	unsigned int tileBase = (tile << 8);
	//unsigned int xoffset;
	
	//baseMemAddress = nonvisiblePage + ((y) << 6) + ((y) << 4);// + (x >> 2);
	baseMemAddress = nonvisiblePage + (y * (scanlineWidthInBytes));// + (x >> 2);
	baseTileAddress = tileBase;
	
	for (cx = 0; cx < 4; cx++)
	{ 
		plane = ((x + cx) & 3);
		/* select the map mask register */
		outp(GAR_INDEX, 0x04);
		/* write 2^plane */
		outp(GAR_DATA, plane);
		
		memAddress = baseMemAddress + ((x + cx) >> 2);
		//memAddress = baseMemAddress;
		tileAddress = baseTileAddress;
		 
		for (cy = 0; cy < 16; cy++)		
		{			
			tiles[tileAddress] = videoMem[memAddress];
			tiles[tileAddress + 4] = videoMem[memAddress + 1];
			tiles[tileAddress + 8] = videoMem[memAddress + 2];
			tiles[tileAddress + 12] = videoMem[memAddress + 3];
			
			tileAddress += 16u;
			memAddress += scanlineWidthInBytes;
		}
		
		baseTileAddress++;	
  
	}		
	

}


void saveTileTo16(int x, int y, unsigned char *destPixels)
{
	int cx;
	int cy;
	//unsigned char pixel;
	int plane;
	unsigned int memAddress;
	unsigned int tileAddress;
	unsigned int baseTileAddress;
	unsigned int baseMemAddress;
	unsigned int tileBase = (unsigned int)destPixels;
	//unsigned int xoffset;
	
	//baseMemAddress = nonvisiblePage + ((y) << 6) + ((y) << 4);// + (x >> 2);
	baseMemAddress = nonvisiblePage + (y * (scanlineWidthInBytes));// + (x >> 2);
	baseTileAddress = tileBase;
	
	for (cx = 0; cx < 4; cx++)
	{ 
		plane = ((x + cx) & 3);
		/* select the map mask register */
		outp(GAR_INDEX, 0x04);
		/* write 2^plane */
		outp(GAR_DATA, plane);
		
		memAddress = baseMemAddress + ((x + cx) >> 2);
		//memAddress = baseMemAddress;
		tileAddress = baseTileAddress;
		 
		for (cy = 0; cy < 16; cy++)		
		{			
			tiles[tileAddress] = videoMem[memAddress];
			tiles[tileAddress + 4] = videoMem[memAddress + 1];
			tiles[tileAddress + 8] = videoMem[memAddress + 2];
			tiles[tileAddress + 12] = videoMem[memAddress + 3];
			
			tileAddress += 16u;
			memAddress += scanlineWidthInBytes;
		}
		
		baseTileAddress++;	
  
	}		
	

}


void copyTile(int sx, int sy, int dx, int dy)
{
	int x;
	int y;
	unsigned char c;
	int srcPos;
	int dstPos;
	
	dx = dx >> 2;	
	dy = dy * scanlineWidthInBytes;
	sx = sx >> 2;
	sy = sy * scanlineWidthInBytes;
	
	x = 0;
	
	srcPos = nonvisiblePage + sx + sy;
	dstPos = nonvisiblePage + dx + dy;

	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);

	
	for (y = 0; y < (16 * scanlineWidthInBytes); y+= scanlineWidthInBytes)  // 1280 = 80 * 16 = height of tile * width of screen
	{			
		c = videoMem[srcPos + y];
		videoMem[dstPos + y] = c;
		c = videoMem[srcPos + y + 1];
		videoMem[dstPos + y + 1] = c;
		c = videoMem[srcPos + y + 2];
		videoMem[dstPos + y + 2] = c;
		c = videoMem[srcPos + y + 3];
		videoMem[dstPos + y + 3] = c;		
	}

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void copyTileFromStorage(int dx, int dy, int tile)
{
	int x;
	int y;
	//unsigned char c;
	int srcPos;
	int dstPos;
	unsigned int scanlineWidthTimes16 = scanlineWidthInBytes << 4;
	
	dx = dx >> 2;	
	dy = dy * scanlineWidthInBytes;
	
	x = 0;
	
	srcPos = videoStorage + ((tile / 20) * scanlineWidthTimes16) + ((tile % 20) << 2);
	dstPos = nonvisiblePage + dx + dy;

	//Enable all planes
	//outp(SC_INDEX, 0x02);  		
	//outp(SC_DATA, 0xFF);
	
	outpw(SC_INDEX, 0xFF02);
	
	//Set bit mask to 0
	//outp(GAR_INDEX, 8);
	//outp(GAR_DATA, 0);
	outpw(GAR_INDEX, 0x0008);

	
	for (y = 0; y < scanlineWidthTimes16; y+= scanlineWidthInBytes)  // 1280 = 80 * 16 = height of tile * width of screen
	{			
		int dst = dstPos + y;
		int src = srcPos + y;
		//videoMem[dst++] = videoMem[src++];		 
		//videoMem[dst++] = videoMem[src++];		
		//videoMem[dst++] = videoMem[src++];		
		//videoMem[dst++] = videoMem[src++];
		
		_asm
		{		
			mov ax, 0xa000
			mov es, ax
			mov di, dst
			mov ds, ax
			mov si, src
			mov cx, 4			
			rep movsb
		};
	}

	//Restore bit mask
	//outp(GAR_INDEX, 8);
	//outp(GAR_DATA, 0xFF);			
	outpw(GAR_INDEX, 0xFF08);
}


void clearColumn16(int column)
{	
	int y;
	int dstPos;
	
	// (16 * scanlineWidthInBytes) + 
	dstPos = nonvisiblePage + (column << 2);

	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
	
	for (y = 0; y < 208; y++)
	{	
		videoMem[dstPos] = 0;	
		videoMem[dstPos + 1] = 0;	
		videoMem[dstPos + 2] = 0;	
		videoMem[dstPos + 3] = 0;
		
		dstPos += scanlineWidthInBytes;
	}
}

void clearRow16(int row)
{	
	int y;
	int dstPos;
	
	// (16 * scanlineWidthInBytes) + 
	dstPos = nonvisiblePage + ((row << 4) * scanlineWidthInBytes);

	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
	
	for (y = 0; y < 16; y++)
	{	
		memset(videoMem + dstPos, 0, scanlineWidthInBytes);
		//for (x = 0; x < scanlineWidthInBytes; x++) videoMem[dstPos + x] = 0;
		dstPos += scanlineWidthInBytes;
	}
}




void shiftScreenLeft(unsigned int srcPos, unsigned int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 208
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		cld
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos
		add si, 4
		mov cx, dx
		sub cx, 4
		rep movs byte ptr es:[di], es:[si]
		
		add dstPos, dx
		add srcPos, dx
		
		pop cx
		loop doloop		
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}




void shiftScreenRight(unsigned int srcPos, unsigned int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage + scanlineWidthInBytes - 4;
	//dstPos = nonvisiblePage + scanlineWidthInBytes - 4;

	srcPos += scanlineWidthInBytes - 4;
	dstPos += scanlineWidthInBytes - 4;

	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 208
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		std
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos
		add di, 3
		dec si
		mov cx, dx
		sub cx, 4
		rep movs byte ptr es:[di], es:[si]
		
		add dstPos, dx
		add srcPos, dx
		
		pop cx
		loop doloop		
		
		cld
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}



void shiftScreenUp(int srcPos, int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	srcPos += (scanlineWidthInBytes << 4);
	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 192
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		cld
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos
		//add si, 4
		mov cx, dx
		//sub cx, 4
		rep movs byte ptr es:[di], es:[si]
		
		add dstPos, dx
		add srcPos, dx
		
		pop cx
		loop doloop		
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void shiftScreenUpLeft(int srcPos, int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	srcPos += (scanlineWidthInBytes << 4);
	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 192
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		cld
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos
		add si, 4
		mov cx, dx
		sub cx, 4
		rep movs byte ptr es:[di], es:[si]
		
		add dstPos, dx
		add srcPos, dx
		
		pop cx
		loop doloop		
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void shiftScreenUpRight(int srcPos, int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;
	
	srcPos += scanlineWidthInBytes - 4 + (scanlineWidthInBytes << 4);
	dstPos += scanlineWidthInBytes - 4;
	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 192
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		std
	  doloop:
		push cx		
		
		mov di, dstPos
		mov si, srcPos
		add di, 3
		dec si
		mov cx, dx
		sub cx, 4
		rep movs byte ptr es:[di], es:[si]
		
		add dstPos, dx
		add srcPos, dx		
		
		pop cx
		loop doloop		
		cld
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void shiftScreenDown(int srcPos, int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	srcPos += scanlineWidthInBytes * 191;
	dstPos += scanlineWidthInBytes * 207;
	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 192
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		cld
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos
		mov cx, dx
		rep movs byte ptr es:[di], es:[si]
		
		sub dstPos, dx
		sub srcPos, dx
		
		pop cx
		loop doloop	
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void shiftScreenDownLeft(int srcPos, int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	srcPos += scanlineWidthInBytes * 191;
	dstPos += scanlineWidthInBytes * 207;
	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 192
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		cld
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos
		add si, 4
		mov cx, dx
		sub cx, 4				
		rep movs byte ptr es:[di], es:[si]
		
		sub dstPos, dx
		sub srcPos, dx
		
		pop cx
		loop doloop	
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void shiftScreenDownRight(int srcPos, int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	srcPos += (scanlineWidthInBytes * 191) + scanlineWidthInBytes - 4;
	dstPos += (scanlineWidthInBytes * 207) + scanlineWidthInBytes - 4;
	
	//srcPos = visiblePage + scanlineWidthInBytes - 4;
	//dstPos = nonvisiblePage + scanlineWidthInBytes - 4;
	
	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 192
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		std
	  doloop:
		push cx
		
		//mov di, dstPos
		//mov si, srcPos
		//mov cx, dx
		//rep movs byte ptr es:[di], es:[si]
		mov di, dstPos
		mov si, srcPos
		add di, 3
		dec si
		mov cx, dx
		sub cx, 4
		rep movs byte ptr es:[di], es:[si]
		
		sub dstPos, dx
		sub srcPos, dx
		
		pop cx
		loop doloop	
		cld
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}


void copyScreen(unsigned int srcPos, unsigned int dstPos)
{
	//int srcPos, dstPos;	
	
	//srcPos = visiblePage;
	//dstPos = nonvisiblePage;

	//Enable all planes
	outp(SC_INDEX, 0x02);  		
	outp(SC_DATA, 0xFF);
			
	//Set bit mask to 0
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0);
		
	_asm {
		
		mov cx, 208
		
		//mov ax, seg scanlineWidthInBytes
		//mov es, ax
		//mov di, offset scanlineWidthInBytes
		//mov dx, word ptr es:[di]
		mov dx, ss:[scanlineWidthInBytes]
	
		mov ax, 0xa000
		mov es, ax
		
		cld
	  doloop:
		push cx
		
		mov di, dstPos
		mov si, srcPos		
		mov cx, dx		
		rep movs byte ptr es:[di], es:[si]
		
		add dstPos, dx
		add srcPos, dx
		
		pop cx
		loop doloop		
	};

	//Restore bit mask
	outp(GAR_INDEX, 8);
	outp(GAR_DATA, 0xFF);			
}




void setVideoMode(int mode)
{
    memset(&regs, 0, sizeof(union REGPACK));
    regs.h.al = mode;
    intr(0x10, &regs);
}


void enableUnchained()
{
	/* turn off chain-4 mode */
	outp(SC_INDEX, MEMORY_MODE);
	outp(SC_DATA, 0x06);

	/* turn off long mode */
	outp(CRTC_INDEX, UNDERLINE_LOC);
	outp(CRTC_DATA, 0x00);
	
	/* turn on byte mode */
	outp(CRTC_INDEX, MODE_CONTROL);
	outp(CRTC_DATA, 0xe3);  //11100011
}



void writePaletteAll(unsigned char startColor, int colorCount, unsigned char *palette)
{ 
	int n;
	
	// mask all registers
	//outp(0x03C6, 0xff);
	
	// set starting color index
	outp(0x03C8, startColor);
	colorCount *= 3;
	
	// Set palette entries
	for (n = 0; n < colorCount; n++ ) outp(0x03C9, palette[n] >> 2);
}

void writePaletteSingle(unsigned char index, unsigned char red, unsigned char green, unsigned char blue)
{ 
	// mask all registers
	//outp(0x03C6, 0xff);
	
	// specify palette index we're setting
	outp(0x03C8, index);
	
	// set palette entry
	outp(0x03C9, red);
	outp(0x03C9, green);
	outp(0x03C9, blue);
}


void readPaletteAll(unsigned char startColor, int colorCount, unsigned char *palette)
{ 
	int n;
	
	// mask all registers
	//outp(0x03C6,0xff);
	
	// specify starting color index
	outp(0x03C7, startColor);
	
	// read palette entries
	for (n = 0; n < colorCount * 3; n++)
	{
		palette[n] = inp(0x03C9) << 2;
	}
}


void readPaletteSingle(unsigned char index, unsigned char *red, unsigned char *green, unsigned char *blue)
{ 
	// mask all registers
	//outp(0x03c6, 0xff);
	
	// specify palette index we're getting
	outp(0x03c7, index);
	
	// get palette entry
	*red = inp(0x03c9);
	*green = inp(0x03c9);
	*blue = inp(0x03c9);
}




void interrupt far kbhandler()
{
	char c;
	char keyDown = 0;
	static char extendedKey = 0;
	
	_asm cli;
	
	//printf("START\n");	
	
	c = inp(0x60);
	//printf("Key: %x\n", c);
	
		
	// upDown == 1 if key down
	keyDown = (c & 0x80) == 0;
	//if (keyDown) printf("DOWN\n");
	
	if (c == 0xe0)
	{
		if (lastChar == 0xaa) { extendedKey = 1; }
	}
	else if (c == 0xe1)
	{
	
	}
	else
	{
		if (lastChar == 0xe0)
		{
			if (extendedKey == 1)
			{
				extendedKey = 0;
			}
			else
			{
				c = c & 0x7F;
				if (c == 0x4b) keyboardLeft = keyDown;
				else if (c == 0x4d) keyboardRight = keyDown;
				else if (c == 0x48) keyboardUp = keyDown;
				else if (c == 0x50) keyboardDown = keyDown;
			}
		}
		else
		{
			c = c & 0x7F;
			if (c == 1 && !keyDown) gameRunning = 0;
			else if (c == 0x31 && !keyDown) noClip = !noClip;
			else if (c == 2 && !keyDown) currentSound = soundHit;
			else if (c == 3 && !keyDown) currentSound = soundFloppy;
			else if (c == 4 && !keyDown) currentSound = soundSteps;
			else if (c == 5 && !keyDown) currentSound = soundAllFloppies;
			else if (c == 6 && !keyDown) currentSound = soundMarker;			
			else if (c == 0x39 && !keyDown) placeMarker = 1;
			//else if (c == 0x13 && !keyDown) drawGrid3D();
			else if (c == 0x1e) keyboardA = keyDown;
			else if (c == 0x20) keyboardD = keyDown;
			else if (c == 0x11) keyboardW = keyDown;
			else if (c == 0x1F) keyboardS = keyDown;
			else if (c == 0x2a) keyboardShiftL = keyDown;
			else if (c == 0x36) keyboardShiftR = keyDown;
		}		
	}
	
	lastChar = c;
	
	keyboardAnyKey = keyDown;
	
	//printf("END\n");
	
	// Acknowledge PIC for interrupt
	outp(0x20, 0x20);
	// Enable interrupts
	_asm sti;		
	
	//oldkbhandler();
}




void drawGrid()
{
	int x, y, c;
	
	for (y = 0; y < 25; y++)
	{
		for (x = 0; x < 40; x++)
		{
			c = grid[(y * 40) + x];			
			if (c > 0) blitTile16(x << 4, y << 4, c - 1);
		}
	}
}


void drawMap()
{
	int x, y, c;
	
	int startX = map_startX;
	int startY = map_startY;
	
	for (x = 0; x < 21; x++)
	{
		for (y = 0; y < 13; y++)
		{
			c = tileMap[((y + startY) << 6) + x + startX]; // '* MAP_WIDTH' replaced with << 6
			if (c == 0) continue;
			c--;
			
			blitTile16(x << 4, (y << 4), c);
		}
	}
}


void drawMapColumn(int columnNum, int xpos, int ypos)
{
	int y, c;	
	
	for (y = 0; y < 13; y++)
	{
		c = tileMap[((y + map_startY) << 6) + columnNum ];  // '<< 6' instead of '* MAP_WIDTH'
		if (c == 0) continue;
		c--;
		
		//blitTile16NoMask(xpos, ypos + (y << 4), c);
		copyTileFromStorage(xpos, ypos + (y << 4), c);
	}
	
}


void drawMapRow(int rowNum, int xpos, int ypos)
{
	int x, c;	
	int basePos = (rowNum << 6) + map_startX;   // '<< 6' instead of '* MAP_WIDTH'
	
	for (x = 0; x < 21; x++)
	{
		c = tileMap[basePos + x];
		if (c == 0) continue;
		c--;
		
		//blitTile16NoMask(xpos + (x << 4), ypos, c);
		copyTileFromStorage(xpos + (x << 4), ypos, c);
	}
	
}


void initMouse()
{
	union REGPACK mregs;
	
	// INT 33h / AX=0000 - mouse ininialization. any previous mouse pointer is hidden. 

	// returns:
	// if successful: AX=0FFFFh and BX=number of mouse buttons.
	// if failed: AX=0
	
	memset(&mregs, 0, sizeof(union REGPACK));
	mregs.w.ax = 0;
	intr(0x33, &mregs);	
	
	
	//  INT 33,F - Set Mouse Mickey Pixel Ratio
	//  AX = 0x0F
	//	CX = horizontal ratio (1..32767, default 8)
	//	DX = vertical ratio (1..32767, default 16)
	//memset(&mregs, 0, sizeof(union REGPACK));
	//mregs.w.ax = 0x0F;
	//mregs.w.cx = 8;
	//mregs.w.dx = 16;
	//intr(0x33, &mregs);
}

void showMouse()
{
	union REGPACK mregs;
	memset(&mregs, 0, sizeof(union REGPACK));
	mregs.w.ax = 1;
	intr(0x33, &mregs);
}

void hideMouse()
{
	union REGPACK mregs;
	memset(&mregs, 0, sizeof(union REGPACK));
	mregs.w.ax = 2;
	intr(0x33, &mregs);
}

void pollMouse()
{
	// INT 33h / AX=0003 - get mouse position and status of its buttons. 

	// returns:
	// if left button is down: BX=1
	// if right button is down: BX=2
	// if both buttons are down: BX=3
	// CX = x
	// DX = y
	
	//union REGPACK mregs;	

	//memset(&mregs, 0, sizeof(union REGPACK));
    //mregs.w.ax = 3;
    //intr(0x33, &mregs);
	
	//mouseX = mregs.w.cx;
	//mouseY = mregs.w.dx;
	//mouseButtons = mregs.w.bx;

	_asm {
		mov ax, 3
		int 0x33
		mov mouseX, cx
		mov mouseY, dx
		mov mouseButtons, bx
	};
	
}


unsigned char hardTileList[] = { 98, 99, 100, 101, 23, 0};
unsigned char hardTileListSize = sizeof(hardTileList);


char isTileHard(int x, int y)
{
	unsigned char _c = tileMap[(y << 6) + x]; // '* MAP_WIDTH' changed to << 6
	int result = 0;
	
	//for (n = 0; n < hardTileListSize; n++) if (hardTileList[n] == _c) return 1;	 
	

	_asm
	{
		xor cx, cx
		//mov dx, seg hardTileListSize
		//mov es, dx
		//mov di, offset hardTileListSize				
		//mov cl, byte ptr es:[di]
		mov cl, byte ptr ss:hardTileListSize
		
		mov dx, seg hardTileList
		mov es, dx
		mov di, offset hardTileList	
	
		mov al, _c
		cld
		repne scasb
		
		or cx, cx		
		jz done
		mov result, 1
		
	  done:		
	
	};
	
	return result;
}



void moveSprite(struct sprite_struct *sprite, int dx, int dy)
{
	int n;	
	if (sprite->x + dx < 0) dx = 0;
	if (sprite->y + dy < 0) dy = 0;		
	
	if (sprite == player && noClip) 
	{ 
		sprite->x += dx;
		sprite->y += dy;
		return;
	}
	
	if (dx != 0 || dy != 0)
	{			
		int newtilex = (sprite->x + dx) >> 4;
		int newtiley = (sprite->y + dy) >> 4;
		
		int tilex = sprite->x >> 4;
		//int tiley = sprite->y >> 4;

		
		if (isTileHard(tilex, newtiley))
		{				
			dy = 0;
		}
		if (((sprite->x) & 15) && isTileHard(tilex + 1, newtiley))
		{				
			dy = 0;
		}
		if (((sprite->y + dy) & 15) && isTileHard(tilex, newtiley + 1))
		{				
			dy = 0;
		}
		if (((sprite->x) & 15) && ((sprite->y + dy) & 15) && isTileHard(tilex + 1, newtiley + 1))
		{				
			dy = 0;
		}
		
		newtiley = (sprite->y + dy) >> 4;
		
		if (isTileHard(newtilex, newtiley))
		{				
			dx = 0;
		}
		if (((sprite->x + dx) & 15) && isTileHard(newtilex + 1, newtiley))
		{				
			dx = 0;
		}
		if (((sprite->y + dy) & 15) && isTileHard(newtilex, newtiley + 1))
		{				
			dx = 0;
		}
		if (((sprite->x + dx) & 15) && ((sprite->y + dy) & 15) && 
				isTileHard(newtilex + 1, newtiley + 1))
		{				
			dx = 0;
		}
		
		for (n = 0; n < MAXSPRITES; n++)
		{
			struct sprite_struct *thissprite = &sprites[n];
			if (dx == 0 && dy == 0) break;			
			
			if (thissprite == sprite || thissprite->type == -1) continue;			
			
			if ((thissprite->y <= sprite->y + dy + 15 && thissprite->y + 15 >= sprite->y + dy) &&
				(thissprite->x <= sprite->x + 15 && thissprite->x + 15 >= sprite->x))
			{
				if ((thissprite == player || sprite == player) && !noClip)
				{
					goToStart = 1;
					break;
				}
				
				dy = 0;
			}
			
			if ((thissprite->x <= sprite->x + dx + 15 && thissprite->x + 15 >= sprite->x + dx) &&
				(thissprite->y <= sprite->y + dy + 15 && thissprite->y + 15 >= sprite->y + dy))
			{
				if ((thissprite == player || sprite == player) && !noClip)
				{
					goToStart = 1;
					break;					
				}
				dx = 0;
			}			
		}
		
		sprite->x += dx;
		sprite->y += dy;
	}

}



char bbVisible(int xpos, int ypos)
{
	if (((xpos >> 4) >= map_startX) && (((xpos + 15) >> 4) <= map_startX + 20) &&
	   ((ypos >> 4) >= map_startY) && (((ypos + 15) >> 4) <= map_startY + 12)) return 1;
	   
	return 0;
}


void generateMap()
{
	unsigned int offsetsX[] = { 0, 1, 0, -1 };
	unsigned int offsetsY[] = { -1, 0, 1, 0 };
	unsigned int opposites[] = { 2, 3, 0, 1 };
	int x, y, n;
	unsigned char mazegrid[100];
	unsigned int mazestack[200];
	int stackposition = 0;	
	int startx = rand() % 10;
	int starty = rand() % 10;
	int lastx = startx;
	int lasty = starty;
	int newx, newy;
	int dir;
	char attempted = 0;
	int highest = 0;
	unsigned char wallTile;
	
	memset(mazegrid, 0, 100);
	
	mazestack[stackposition] = startx | (starty << 8);
	
	// Generate 10x10 maze
	while (stackposition >= 0)
	{
		int thisx = mazestack[stackposition] & 0xFF;
		int thisy = (mazestack[stackposition] >> 8) & 0xFF;
		char thismaze = mazegrid[(thisy * 10) + thisx];		
		
		
		attempted = 0;		
		while (attempted != 15)
		{
			newx = newy = -1;			
			dir = rand() % 4;
			
			if (attempted & (1 << dir)) continue;			
			attempted |= (1 << dir);
			if (thismaze & (1 << dir)) continue;
			
			newx = thisx + offsetsX[dir];
			newy = thisy + offsetsY[dir];
			if (newx < 0 || newy < 0 || newx >= 10 || newy >= 10) { newx = newy = -1; continue; }
			if (mazegrid[(newy * 10) + newx] != 0) { newx = newy = -1; continue; }
			
			break;			
		}
		
		if (newx == -1 && newy == -1)
		{			
			stackposition--;
			continue;
		}		
		
		mazegrid[(thisy * 10) + thisx] |= (1 << dir);
		mazegrid[(newy * 10) + newx] |= (1 << opposites[dir]);
		
		mazestack[++stackposition] = newx | (newy << 8);	
		
		if (stackposition > highest) { highest = stackposition; maze_endX = (newx * 6) + 3; maze_endY = (newy * 6) + 3; }
		
	}
	
	
	maze_startX = ((startx * 6) + 3)* 16;
	maze_startY = ((starty * 6) + 3)* 16;
	
	
	wallTile = 98 + (currentLevel & 3);
	// Clear out existing tile map
	memset(tileMap, TILE_FLOOR, 64 * 64);
	for (n = 0; n < 64; n++) 
	{ 		
		tileMap[n] = wallTile; 
		tileMap[(63 * 64) + n] = wallTile; 
		tileMap[n * 64] = wallTile;  
		tileMap[(n * 64) + 63] = wallTile; 
	}	

	// Generate tiles based on maze
	for (y = 0; y < 10; y++)
	{
		for (x = 0; x < 10; x++)
		{
			int tx, ty;
			//printf("%d ", mazegrid[(y * 10) + x]);
			
			// if ((mazegrid[(y * 10) + x] & 1) == 0)
			// {			
				// for (tx = 0; tx < 6; tx++)
				// {
					// int tilex = (x * 6) + 1 + tx;
					// int tiley = (y * 6) + 1;
					// tileMap[(tiley * MAP_WIDTH) + tilex] = 98;
				// }
			// }

			if ((mazegrid[(y * 10) + x] & 4) == 0)
			{			
				for (tx = 0; tx < 7; tx++)
				{
					int tilex = (x * 6) + tx;
					int tiley = (y * 6) + 1 + 5;
					tileMap[(tiley * MAP_WIDTH) + tilex] = wallTile;
				}
			}
			
			if ((mazegrid[(y * 10) + x] & 2) == 0)
			{
				for (ty = 0; ty < 7; ty++)
				{
					int tilex = (x * 6) + 1 + 5;
					int tiley = (y * 6) + ty;
					tileMap[(tiley * MAP_WIDTH) + tilex] = wallTile;
				}
			}
			
			// if ((mazegrid[(y * 10) + x] & 8) == 0)
			// {
				// for (ty = 0; ty < 6; ty++)
				// {
					// int tilex = (x * 6) + 1;
					// int tiley = (y * 6) + 1 + ty;
					// tileMap[(tiley * MAP_WIDTH) + tilex] = 98;
				// }
			// }		
		}
		//printf("\n");
	}
	
	tileMap[(maze_endY << 6) + maze_endX] = 81;
	
	// Draw shadow tiles
	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			int tile = tileMap[(y << 6) + x];
			if (tile == TILE_FLOOR)
			{
				if (tileMap[((y-1) << 6) + x] == wallTile)
				{
					// corner bracket
					if (tileMap[((y) << 6) + x + 1] == wallTile) tileMap[((y) << 6) + x] = 120;
					// horizontal side
					else if (tileMap[((y-1) << 6) + x + 1] == wallTile) tileMap[((y) << 6) + x] = 152;
					// horizontal end
					else tileMap[((y) << 6) + x] = 153;
					
				}
				else if (tileMap[((y) << 6) + x + 1] == wallTile)
				{
					// vertical side
					if (tileMap[((y-1) << 6) + x + 1] == wallTile) tileMap[((y) << 6) + x] = 118;
					// vertical end
					else tileMap[((y) << 6) + x] = 102;
				}
				// corner square
				else if (tileMap[((y-1) << 6) + x + 1] == wallTile) tileMap[((y) << 6) + x] = 150;
			}
		}
	}
	
	// Generate floppy disks
	for (n = 0; n < 10; n++)
	{
		x = rand() % 60;
		y = rand() % 60;
		
		if (tileMap[(y << 6) + x] == TILE_FLOOR)
		{
			tileMap[(y << 6) + x] = TILE_FLOPPYDISK;
		}
		else n--;
		
	}
	
}


void positionRobots(char setLast)
{
	int c;
	
	for (c = MAXSPRITES-1; c > 0; c--)
	{
		sprites[c].type = 1;
		do
		{
			sprites[c].x = (rand() % 64) << 4;
			sprites[c].y = (rand() % 64) << 4;
			if (setLast)
			{
				sprites[c].lastx = sprites[c].x;
				sprites[c].lasty = sprites[c].y;
			}
		}
		while (isTileHard(sprites[c].x >> 4, sprites[c].y >> 4));
		
		sprites[c].tile = 52;
		if (rand() % 1) { sprites[c].xvel = (rand() % 1 ? -ROBOT_SPEED : ROBOT_SPEED); sprites[c].yvel = 0; }
		else { sprites[c].yvel = (rand() % 1 ? -ROBOT_SPEED : ROBOT_SPEED); sprites[c].xvel = 0; }
		//if (bbVisible(sprites[c].x, sprites[c].y))
		{
			//saveTileTo16(sprites[c].x - (map_startX << 4), sprites[c].y - (map_startY << 4), sprites[c].background);
			//blitTile16(sprites[c].x - (map_startX << 4), sprites[c].y - (map_startY << 4), sprites[c].tile);
		}
	}
	

}


void updateStatusBar()
{
	int n;
	
	tempPage = nonvisiblePage;
	nonvisiblePage = 0;
	
	// clear floppies
	for (n = 0; n < 10; n++)
	{
		blitTile16(2 + (n * 10), 2, 177);
	}
	
	for (n = 0; n < 10; n++)
	{
		if ((10 - floppiesCollected) > n) blitTile16(2 + (n * 10), 2, 176);
	}
	
	nonvisiblePage = tempPage;
}



void titleScreen()
{
	int xoffset = 0;
	int yoffset = 4;
	int c;
	char story[] = "MORTIMER A. ZEMAN";
	char titleFade = 0;
	
	keyboardAnyKey = 0;		
	
	// select the map mask register
	outp(SC_INDEX, 0x02);
	// enable all planes
	outp(SC_DATA, 0xFF);
	// clear video mem
	//memset(videoMem + visiblePage, 20, scanlineWidthInBytes * 208u);
	//memset(videoMem + nonvisiblePage, 20, scanlineWidthInBytes * 208u);
	
	// for (c = 0; c < strlen(story); c++)
	// {
		// blitTile16(c * 12, 0, asciiTiles[story[c]]  - 1);
	// }
	
	// Swap page offsets		
	tempPage = visiblePage;
	visiblePage = nonvisiblePage;
	nonvisiblePage = tempPage;
	
	page_addr = visiblePage + (yoffset * scanlineWidthInBytes) + (xoffset >> 2);
	high_addr = HIGH_ADDRESS | (page_addr & 0xff00);
	low_addr = LOW_ADDRESS  | (page_addr << 8);

	
	while (1)
	{
		if (keyboardAnyKey) titleFade = 1;
		
		if (titleFade)
		{
			int n;
			int c = 0;
			for (n = 0; n < 256 * 3; n++)
			{
				if (currentPalette[n] > 16) { currentPalette[n] -= 16; c = 1; }
				else currentPalette[n] = 0;
				//currentPalette[n] = currentPalette[n] >> 1;				
			}
			if (!c) titleFade = 2;
		}
	
		// wait for previous trace to finish
		//_asm cli
		while (inp(0x3DA) & 0x8) {}	
		
		// switch starting video address
		outpw(CRTC_INDEX, high_addr);
		outpw(CRTC_INDEX, low_addr);	
		
		// set horizontal pixel shift
		inp(0x3DA);
		outp(ACR_INDEX, 0x33); //0x13, 0x33
		outp(ACR_INDEX, (xoffset % 4) << 1);	
		
		if (titleFade) writePaletteAll(0, 256, currentPalette);
		
		while (!(inp(0x3DA) & 0x8)) {}	
		
		
		if (titleFade == 2) break;
	}
	
	while (keyboardAnyKey == 1) {}
	placeMarker = 0;
}


void showLevelScreen()
{
	int xoffset = 0;
	int yoffset = 0;
	int c;	
	char titleFade = 0;
	char levelText[] = "LEVEL";
	
	keyboardAnyKey = 0;		
	
	// select the map mask register
	outp(SC_INDEX, 0x02);
	// enable all planes
	outp(SC_DATA, 0xFF);
	// clear video mem
	memset(videoMem + visiblePage, 0, scanlineWidthInBytes * 208u);
	memset(videoMem + nonvisiblePage, 0, scanlineWidthInBytes * 208u);
	
	for (c = 0; c < strlen(levelText); c++)
	{
		blitTile16((c * 12) + (8 * 16), 4 * 16, asciiTiles[levelText[c]]  - 1);
	}
	sprintf(levelText, "%02d", currentLevel+1);
	for (c = 0; c < strlen(levelText); c++)
	{
		blitTile16((c * 12) + 3 + (9 * 16), 5 * 16, asciiTiles[levelText[c]]  - 1);
	}
	
	// Swap page offsets		
	tempPage = visiblePage;
	visiblePage = nonvisiblePage;
	nonvisiblePage = tempPage;
	
	page_addr = visiblePage + (yoffset * scanlineWidthInBytes) + (xoffset >> 2);
	high_addr = HIGH_ADDRESS | (page_addr & 0xff00);
	low_addr = LOW_ADDRESS  | (page_addr << 8);

	
	memcpy(currentPalette, tiles_palette, 256 * 3);
	while (inp(0x3DA) & 0x8) {}		
	writePaletteAll(0, 256, currentPalette);
	while (!(inp(0x3DA) & 0x8)) {}	
	
	
	while (1)
	{
		if (keyboardAnyKey) titleFade = 2;
		
		if (titleFade)
		{
			int n;
			int c = 0;
			for (n = 0; n < 256 * 3; n++)
			{
				if (currentPalette[n] > 16) { currentPalette[n] -= 16; c = 1; }
				else currentPalette[n] = 0;
				//currentPalette[n] = currentPalette[n] >> 1;				
			}
			if (!c) titleFade = 2;
		}
	
		// wait for previous trace to finish
		//_asm cli
		while (inp(0x3DA) & 0x8) {}	
		
		// switch starting video address
		outpw(CRTC_INDEX, high_addr);
		outpw(CRTC_INDEX, low_addr);	
		
		// set horizontal pixel shift
		inp(0x3DA);
		outp(ACR_INDEX, 0x33); //0x13, 0x33
		outp(ACR_INDEX, (xoffset % 4) << 1);	
		
		if (titleFade) writePaletteAll(0, 256, currentPalette);
		
		while (!(inp(0x3DA) & 0x8)) {}	
		
		
		if (titleFade == 2) break;
	}
	
	placeMarker = 0;
	
	while (keyboardAnyKey == 1) {}
	placeMarker = 0;
}



int main(int args, char* argv[])
{
    int x;
    int y;
	//int z;
    int c;
	clock_t lastTime;
	clock_t thisTime;
	int deltaTime;
	int fpsCounter;
	int fpsTimer;
	int fps;
	int xoffset = 0;
	int yoffset = 0;
	int lastxoffset = 0;
	int lastyoffset = 0;
	//char screenDrawn = 0;
	//signed char moveDirX = 1;
	//signed char moveDirY = 1;
    
    printf("Hello world!\n");
	
	
	

	//generateMap();
	//exit(0);
	
	//while (!mouseButtons) { pollMouse();	 printf("X: %d Y: %d\n", mouseX, mouseY); }
	//exit(0);
	
	soundOff();
	
	// Replace keyboard interrupt handler
	oldkbhandler = _dos_getvect(9);
	_dos_setvect(9, &kbhandler);	
	
	
	
	// Disable interrupts
	//_asm cli;	
	//while (inp(0x64) & 2) inp(0x60); // Empty keyboard	
	//while (inp(0x64) & 2) {} // wait until can write
	//outp(0x64, 0xAD);   // disable keyboard
	//while (inp(0x64) & 2) {}
	//outp(0x60, 0xF8);	// Disable typematic mode (autorepeat)
	//while ((inp(0x64) & 2) == 0) {}
	//printf("Result: %x\n", inp(0x60));
	//while (inp(0x64) & 2) {}
	//outp(0x64, 0xAE);   // enable keyboard	
	// Enable interrupts
	//_asm sti;
	
	
	//while (gameRunning) {}	
	//_dos_setvect(9, oldkbhandler);
	//return 0;
	

	setVideoMode(0x13);
	enableUnchained();

	initMouse();	
	
	//showMouse();

	// set up timing variables
	fpsCounter = 0;
	fpsTimer = 0;
	fps = 0;	
	lastTime = clock();
	
	scanlineWidthInBytes = (320 + 16) / 4;
	
	videoMem = (unsigned char far*) MK_FP(0xA000,0x0000);
	visiblePage = scanlineWidthInBytes * 12u;  // 8u
	nonvisiblePage = visiblePage + scanlineWidthInBytes * 208u;
	videoStorage = nonvisiblePage + scanlineWidthInBytes * 208u;
	
	
	//nonvisiblePage = visiblePage;
	
	// select the map mask register
	outp(SC_INDEX, 0x02);
	// enable all planes
	outp(SC_DATA, 0xFF);
	// clear video memory
	memset(videoMem, 0, scanlineWidthInBytes * 8u);
	memset(videoMem + visiblePage, 0, scanlineWidthInBytes * 208u);
	memset(videoMem + nonvisiblePage, 0, scanlineWidthInBytes * 208u);
	memset(videoMem + videoStorage, 0, scanlineWidthInBytes * 208u);

		
	// Set width of rows for panning
	outp(CRTC_INDEX, 0x13);
	outp(CRTC_DATA, scanlineWidthInBytes >> 1);	
	

	
	// Set Pixel Panning Mode to 0 (keeps status bar still)
	inp(0x3DA);
	outp(ACR_INDEX, 0x10);
	c = inp(ACR_DATA);
	inp(0x3DA);
	outp(ACR_INDEX, 0x10);
	outp(ACR_INDEX, c | 0x20);
	
	
	// Put tiles in video memory for fast blitting
	x = 0;
	y = 0;
	tempPage = nonvisiblePage;
	nonvisiblePage = videoStorage;
	for (c = 0; c < 256; c++)
	{
		blitTile16(x << 4, y << 4, c);
		x++;
		if (x >= 20) { x = 0; y++; }
	}
	nonvisiblePage = tempPage;
	
	
	srand(clock() + time(NULL));		
	
	
	for (c = 0; c < 26; c++)
	{
		if (c < 10) asciiTiles[48 + c] = c + 208 + 1;
		asciiTiles[c + 65] = c + 224 + 1;
	}
	asciiTiles[32] = 193;
	asciiTiles[','] = 219;
	asciiTiles['\''] = 220;
	asciiTiles['('] = 221;
	asciiTiles[')'] = 222;
	asciiTiles['.'] = 223;
	
	
	

	// print some test text
	//for (c = 0; c < strlen(copyright); c++)
	//{
	//	tileMap[c] = asciiTiles[copyright[c]];
	//}
	
	//drawGrid();
	//drawMap();	
	
	
	
	// Save original palette
	readPaletteAll(0, 256, originalPalette);
	// Setup current palette
	memcpy(currentPalette, tiles_palette, 256 * 3);
	writePaletteAll(0, 256, currentPalette);
	
	xoffset = 0;
	yoffset = 0;
	
	// Set up initial mouse state and get mask tile
	//pollMouse();
	//saveTile16(xoffset + (mouseX>>1), yoffset + mouseY, 33);		
	//mouseLastX = xoffset + (mouseX>>1);
	//mouseLastY = yoffset + mouseY;
	//mouseLastButtons = mouseButtons;
	
	
		
	for (c = 0; c < MAXSPRITES; c++)
	{
		memset(&sprites[c], 0, sizeof(struct sprite_struct));
		sprites[c].type = -1;
	}
	
	
	
	
	frameCounter = 0;

	map_startX = 0;
	map_startY = 0;
	drawMap();	
	titleScreen();
	

	// Setup status bar area at bottom of screen
	//
	// Maximum Scan Line Register
	// unset bit 9 of Line Compare Register
	outp(CRTC_INDEX, 0x09);
	c = inp(CRTC_DATA);
	outp(CRTC_INDEX, 0x09);
	outp(CRTC_DATA, c & 0xBF);		
	//Overflow Register
	// set bit 8 of Line Compare Register
	outp(CRTC_INDEX, 0x07);
	c = inp(CRTC_DATA);
	outp(CRTC_INDEX, 0x07);
	outp(CRTC_DATA, c | 0x10);	
	//Line Compare Register (Index 18h)
	outp(CRTC_INDEX, 0x18);
	outp(CRTC_DATA, 119);	// 111 for 16-pixel status bar, 127 for 8-pixel

	
	
	
	
	generateMap();			
	
	positionRobots(1);	
	
	player = &sprites[0];
	
	player->type = 0;
	player->x = player->lastx = maze_startX;
	player->y = player->lasty = maze_startY;	
	player->tile = 39;
	player->dir = 1;
	//saveTileTo16(player->x - (map_startX << 4), player->y - (map_startY << 4), player->background);
	//blitTile16(player->x - (map_startX << 4), player->y - (map_startY << 4), player->tile);
	
	map_startX = (maze_startX >> 4) - 10;
	map_startY = (maze_startY >> 4) - 6;
	if (map_startX < 0) map_startX = 0;
	if (map_startY < 0) map_startY = 0;
	if (map_startX + 20 >= MAP_WIDTH - 3) { map_startX = MAP_WIDTH - 20 - 3; }  // -3 for unused maze area
	if (map_startY + 12 >= MAP_HEIGHT - 3) { map_startY = MAP_HEIGHT - 12 - 3; }
	
	map_startX = 32;
	map_startY = 32;	
	//drawMap();	

	
	//tempPage = visiblePage;
	//visiblePage = nonvisiblePage;
	//nonvisiblePage = tempPage;
	
	updateStatusBar();

	logicEnabled = 0;
	manualScrollTimer = 20;	
	
	showLevelScreen();
	
	
	drawMap();		
	tempPage = visiblePage;
	visiblePage = nonvisiblePage;
	nonvisiblePage = tempPage;
	
	memcpy(currentPalette, tiles_palette, 256 * 3);
	while (inp(0x3DA) & 0x8) {}		
	writePaletteAll(0, 256, currentPalette);
	while (!(inp(0x3DA) & 0x8)) {}	

	
	/* * * * * * * * * 
	 *
	 * MAIN LOOP 
	 *
	 */
	while (gameRunning)	
    {
		int sprite_dx;
		int sprite_dy;
		char shiftLeft;
		char shiftRight;
		char shiftUp;
		char shiftDown;
	
	
		thisTime = clock();
		deltaTime = thisTime - lastTime;
		lastTime = thisTime;
		
		fpsTimer += deltaTime;
		if (fpsTimer > 1000)
		{
			fpsTimer -= 1000;
			fps = fpsCounter;
			fpsCounter = 0;
		}
		fpsCounter++;		 

		
		if (goNextLevel)
		{
			if (goNextLevel == 1)
			{
				logicEnabled = 0;
				currentSound = soundSteps;				
				goNextLevel = 2;
			}
			else if (goNextLevel == 2)
			{
				if (currentSound == NULL) { fadeout = 1; goNextLevel = 3; }
			}
			else if (goNextLevel == 3)
			{		
				if (fadeout == 2 && currentSound == NULL)
				{
					memcpy(currentPalette, tiles_palette, 256 * 3);
					while (inp(0x3DA) & 0x8) {}		
					writePaletteAll(0, 256, currentPalette);
					while (!(inp(0x3DA) & 0x8)) {}	
					
					fadeout = 3;		
					
					currentLevel++;
					showLevelScreen();					
					
					generateMap();
					positionRobots(1);
					player->x = player->lastx = maze_startX;
					player->y = player->lasty = maze_startY;
					map_startX = 32;
					map_startY = 32;
					drawMap();
					frameCounter = 1;
					logicEnabled = 1;
					floppiesCollected = 0;
					goNextLevel = 0;
					updateStatusBar();
				}			
			}
		}
		
		
		if (goToStart)
		{
			if (goToStart == 1)
			{
				currentSound = soundHit;
				player->x = maze_startX;
				player->y = maze_startY;
				logicEnabled = 0;
				positionRobots(0);
				
				goToStart = 2;
			}
			else if (goToStart == 2)
			{
				goToStart = 0;
			}
		}
		
		
		if (currentSound != NULL)
		{
			if (*currentSound == 0) 
			{ 
				soundOff(); 
				currentSound = NULL; 
			}
			else if (*currentSound == -1) { currentSound++; }
			else if (*currentSound == -2) { soundOff(); currentSound++; }
			else
			{
				soundOn(*currentSound);
				currentSound++;
			}
		}
		
		

		if (fadeout == 1)
		{
			int n;
			int c = 0;
			for (n = 0; n < 256 * 3; n++)
			{
				if (currentPalette[n] > 16) { currentPalette[n] -= 16; c = 1; }
				else currentPalette[n] = 0;
				//currentPalette[n] = currentPalette[n] >> 1;				
			}
			if (!c) fadeout = 2;
		}		
		
		
		
		
		#define MOVESPEED 4
		#define MOVESPEED2 8
		#define KEEPMOVING 1
		
		if ((frameCounter & 1) == 0)
		{
			//if (player->x != player->lastx || player->y != player->lasty)
			//{
				//if (player->x >= 160 && player->x < 860 - 48) xoffset += player->x - player->lastx;				
				//if (player->y >= 96 && player->y < 924 - 48) yoffset +=  player->y - player->lasty;
			//}
			

		

			if (manualScrollTimer < 20) manualScrollTimer++;
			else			
			{			
				char lockedX = 1;
				char lockedY = 1;
				
				if (player->x < ((map_startX << 4) + xoffset) + 160)
				{
					int diff = (((map_startX << 4) + xoffset) + 160) - player->x;
					if (diff >= MOVESPEED2) { xoffset -= MOVESPEED2; lockedX = 0; }
					else if (keyboardShiftL && (diff >= (SPRITE_SPEED << 1))) xoffset -= (SPRITE_SPEED << 1);
					else xoffset -= SPRITE_SPEED;
				}
				else if (player->x > ((map_startX << 4) + xoffset) + 160) 
				{
					int diff = player->x - (((map_startX << 4) + xoffset) + 160);
					if (diff >= MOVESPEED2) { xoffset += MOVESPEED2; lockedX = 0; }
					else if (keyboardShiftL && (diff >= (SPRITE_SPEED << 1))) xoffset += (SPRITE_SPEED << 1);
					else xoffset += SPRITE_SPEED; 		
				}		
				
				if (player->y < ((map_startY << 4) + yoffset) + 96)
				{
					int diff = (((map_startY << 4) + yoffset) + 96) - player->y;
					if (diff >= MOVESPEED2) { yoffset -= MOVESPEED2; lockedY = 0; }
					else if (keyboardShiftL && (diff >= (SPRITE_SPEED << 1))) yoffset -= (SPRITE_SPEED << 1);
					else yoffset -= SPRITE_SPEED; 
				}
				else if (player->y > ((map_startY << 4) + yoffset) + 96) 
				{
					int diff = player->y - (((map_startY << 4) + yoffset) + 96);
					if (diff >= MOVESPEED2) { yoffset += MOVESPEED2; lockedY = 0; }
					else if (keyboardShiftL && (diff >= (SPRITE_SPEED << 1))) yoffset += (SPRITE_SPEED << 1);
					else yoffset += SPRITE_SPEED; 
				}
				
				if (xoffset == lastxoffset) lockedX = 1;
				if (yoffset == lastyoffset) lockedY = 1;
				
				if (lockedX && lockedY && (goNextLevel == 0) && (goToStart == 0)) logicEnabled = 1;
				//if (lockedX && lockedY && skipRender) skipRender = 0;
			}
			
			
			lastxoffset = xoffset;
			lastyoffset = yoffset;
			
			if (keyboardLeft) 
			{ 
				manualScrollTimer = 0; 
				xoffset -= (keyboardShiftL == 0 ? MOVESPEED : MOVESPEED2); 
				keyboardLeft = KEEPMOVING; 				
			}
			if (keyboardRight) 
			{ 
				manualScrollTimer = 0; 
				xoffset += (keyboardShiftL == 0 ? MOVESPEED : MOVESPEED2); 
				keyboardRight = KEEPMOVING; 				
			}
			if (keyboardDown) 
			{ 
				manualScrollTimer = 0; 
				yoffset += (keyboardShiftL == 0 ? MOVESPEED : MOVESPEED2); 
				keyboardDown = KEEPMOVING; 				
			}
			if (keyboardUp) 
			{ 
				manualScrollTimer = 0; 
				yoffset -= (keyboardShiftL == 0 ? MOVESPEED : MOVESPEED2); 
				keyboardUp = KEEPMOVING; 				
			}
			
			//xoffset+= 4;
			

			
			shiftUp = shiftDown = shiftLeft = shiftRight = 0;
			
			//copyScreenV2NV();	

			if (xoffset >= 16) 
			{ 		
				//xoffset = 16;
				map_startX++;
				if (map_startX + 20 >= MAP_WIDTH - 3) { map_startX = MAP_WIDTH - 21 - 3; xoffset = 16; } //copyScreenV2NV(); } // -3 for unused tiles for maze
				else
				{
					//blitTile16(mouseLastX, mouseLastY, 33);					
					while (xoffset >= 16) xoffset -= 16;
					////shiftScreenLeftV2NV(); 		
					//clearColumn16(20);
					////drawMapColumn(map_startX + 20, 320, 0);
					shiftLeft = 1;
					//saveTile16(xoffset + (mouseX>>1), yoffset + mouseY, 33);		
				}
			}
			else if (xoffset < 0) 
			{ 			
				map_startX--;
				if (map_startX < 0) { map_startX = 0; xoffset = 0; } //copyScreenV2NV(); }
				else
				{
					//blitTile16(mouseLastX, mouseLastY, 33);					
					while (xoffset < 0) xoffset += 16;
					//xoffset = 15; 			
					////shiftScreenRightV2NV(); 
					//clearColumn16(0);
					////drawMapColumn(map_startX, 0, 0);
					shiftRight = 1;
					//saveTile16(xoffset + (mouseX>>1), yoffset + mouseY, 33);		
				}			
			}
			else
			{
				
				////copyScreenV2NV();
			}
			
			if (yoffset >= 16)
			{			
				map_startY++;
				if (map_startY + 12 >= MAP_HEIGHT - 3) { map_startY = MAP_HEIGHT - 13 - 3; yoffset = 16; } // -3 for unused tiles for maze
				else
				{
					//yoffset = 0;			
					while (yoffset >= 16) yoffset -= 16;
					////shiftScreenUp(nonvisiblePage, nonvisiblePage); 
					//clearRow16(12);
					////drawMapRow(map_startY + 12, 0, 192);
					shiftUp = 1;
				}
			}
			else if (yoffset < 0)
			{			
				map_startY--;
				if (map_startY < 0) { map_startY = 0; yoffset = 0; }
				else
				{
					//yoffset = 15;				
					while (yoffset < 0) yoffset += 16;
					////shiftScreenDown(nonvisiblePage, nonvisiblePage); 
					//clearRow16(0);
					////drawMapRow(map_startY, 0, 0);
					shiftDown = 1;
				}
			}
			
			
			
			if (shiftLeft && shiftUp)  // move down-right
			{
				//shiftScreenLeftV2NV(); 		
				//shiftScreenUp(nonvisiblePage, nonvisiblePage); 
				shiftScreenUpLeft(visiblePage, nonvisiblePage); 
				drawMapColumn(map_startX + 20, 320, 0);
				drawMapRow(map_startY + 12, 0, 192);			
			}
			else if (shiftRight && shiftUp)
			{
				//shiftScreenRightV2NV(); 
				//shiftScreenUp(nonvisiblePage, nonvisiblePage); 
				shiftScreenUpRight(visiblePage, nonvisiblePage); 
				drawMapColumn(map_startX, 0, 0);
				drawMapRow(map_startY + 12, 0, 192);
			}
			else if (shiftLeft && shiftDown)
			{
				//shiftScreenLeftV2NV(); 	
				//shiftScreenDown(nonvisiblePage, nonvisiblePage); 
				shiftScreenDownLeft(visiblePage, nonvisiblePage); 
				drawMapColumn(map_startX + 20, 320, 0);			
				drawMapRow(map_startY, 0, 0);
			}
			else if (shiftRight && shiftDown)
			{
				//shiftScreenRightV2NV(); 
				//shiftScreenDown(nonvisiblePage, nonvisiblePage); 
				shiftScreenDownRight(visiblePage, nonvisiblePage); 
				drawMapColumn(map_startX, 0, 0);
				drawMapRow(map_startY, 0, 0);
			}		
			else if (shiftLeft)
			{
				shiftScreenLeft(visiblePage, nonvisiblePage);
				drawMapColumn(map_startX + 20, 320, 0);
			}
			else if (shiftRight)
			{
				shiftScreenRight(visiblePage, nonvisiblePage); 
				drawMapColumn(map_startX, 0, 0);
			}		
			else if (shiftUp)
			{
				//copyScreenV2NV();	
				shiftScreenUp(visiblePage, nonvisiblePage); 
				drawMapRow(map_startY + 12, 0, 192);
			}
			else if (shiftDown)		
			{
				//copyScreenV2NV();	
				shiftScreenDown(visiblePage, nonvisiblePage); 
				drawMapRow(map_startY, 0, 0);
			}
			else copyScreen(visiblePage, nonvisiblePage);	
		}
		
		
		screen_xpixeloffset = (map_startX << 4);
		screen_ypixeloffset = (map_startY << 4);
		
		
		
		
		tileUpdatePosition = 0;
		
		//player.x++;
		sprite_dx = 0;
		sprite_dy = 0;
		if (keyboardA) 
		{ 
			sprite_dx = -SPRITE_SPEED; 
			if (player->dir != 3)
			{
				player->dir = 3; 
				player->tile = 36;
				player->animCounter = 0;
			}
			keyboardA = 1; 
			keyboardLast |= 8;
		}
		if (keyboardD) 
		{ 
			sprite_dx = SPRITE_SPEED; 
			if (player->dir != 1)
			{
				player->dir = 1; 
				player->tile = 39;
				player->animCounter = 0;				
			}
			keyboardD = 1; 
			keyboardLast |= 2;
		}
		if (keyboardW) { sprite_dy = -SPRITE_SPEED; keyboardW = 1; keyboardLast |= 1; }
		if (keyboardS) { sprite_dy =  SPRITE_SPEED; keyboardS = 1; keyboardLast |= 4; }
		
		if (keyboardShiftL != 0) { sprite_dx *= 2; sprite_dy *= 2; }
		
		

		player->lastxvel = player->xvel;
		player->lastyvel = player->yvel;
		player->xvel = sprite_dx;
		player->yvel = sprite_dy;
		
		//if ((player->lastxvel != sprite_dx) || (player->lastyvel != sprite_dy)) skipRender = 1; else skipRender = 0;
		

		//frameCounter++;
		
		if ((frameCounter & 1) == 1)
		{
		
			if (logicEnabled)
			for (c = 0; c < MAXSPRITES; c++)
			{
				struct sprite_struct *thissprite = &sprites[c];
				if (thissprite->type == -1) continue;
				
				thissprite->lastx = thissprite->x; 
				thissprite->lasty = thissprite->y; 
				
				
				moveSprite(thissprite, thissprite->xvel, thissprite->yvel);
				
				if (c != 0)
				{
					// if (thissprite->x == thissprite->lastx)
					// {
						// if (thissprite->xvel == 1) thissprite->xvel = -1;
						// else thissprite->xvel = 1;
					// }
					// if (thissprite->y == thissprite->lasty)
					// {
						// if (thissprite->yvel == 1) thissprite->yvel = -1;
						// else thissprite->yvel = 1;
					// }
					
					if ((thissprite->x == thissprite->lastx && thissprite->y == thissprite->lasty) || (rand() % 1000) == 79)
					{
						int newdir = rand() % 4;
						switch (newdir)
						{
							case 0: thissprite->xvel = 0; thissprite->yvel = -ROBOT_SPEED; break;
							case 1: thissprite->xvel = ROBOT_SPEED; thissprite->yvel = 0; thissprite->dir = 1; break;
							case 2: thissprite->xvel = 0; thissprite->yvel = ROBOT_SPEED; break;
							case 3: thissprite->xvel = -ROBOT_SPEED; thissprite->yvel = 0; thissprite->dir = 3; break;
						}						
					}
				}
				
				
				// Check for player collisions with special tiles
				if (c == 0 && ((thissprite->x != thissprite->lastx) || (thissprite->y != thissprite->lasty)))
				{
					int tilex;
					int tiley;
					int n;
					
					for (n = 0; n < 4; n++)
					{					
						switch (n)
						{
							case 0: tilex = thissprite->x >> 4;	tiley = thissprite->y >> 4; break;
							case 1: tilex = (thissprite->x + 15) >> 4;	tiley = thissprite->y >> 4; break;
							case 2: tilex = thissprite->x >> 4;	tiley = (thissprite->y + 15) >> 4; break;
							case 3: tilex = (thissprite->x + 15) >> 4;	tiley = (thissprite->y + 15) >> 4; break;
						}
						if (tileMap[(tiley << 6) + tilex] == TILE_FLOPPYDISK)
						{
							currentSound = soundFloppy;
							tileMap[(tiley << 6) + tilex] = TILE_FLOOR;
							floppiesCollected++;
							if (floppiesCollected == 10) currentSound = soundAllFloppies;
							updateStatusBar();
						}
						else if (tileMap[(tiley << 6) + tilex] == TILE_EXIT)
						{
							int midx = player->x + 8;
							int midy = player->y + 8;
							int tx = tilex << 4;
							int ty = tiley << 4;
							if (midx >= (tx + 4) && midx <= (tx + 11) && midy >= (ty + 4) && midy <= (ty + 11))
							{						
								// Fade out to go to next level								
								goNextLevel = 1;
							}
						}
					}
				}		

				if (c == 0 && placeMarker)
				{
					int markerX = (player->x + 8) >> 4;
					int markerY = (player->y + 8) >> 4;
					if (tileMap[(markerY << 6) + markerX] == TILE_FLOOR)					
						tileMap[(markerY << 6) + markerX] = TILE_FLOPPYDISK + 1;
					else if (tileMap[(markerY << 6) + markerX] == TILE_FLOPPYDISK + 1)
						tileMap[(markerY << 6) + markerX] = TILE_FLOOR;
					currentSound = soundMarker;
					placeMarker = 0;
				}
				
				// Update player animation				
				if (thissprite->type == 0)
				{					
					if (thissprite->xvel || thissprite->yvel) thissprite->animCounter++;
					if (thissprite->animCounter > 2)
					{
						thissprite->animCounter = 0;
						thissprite->tile++;
						//if (thissprite->tile > 68) thissprite->tile = 64;
						if (thissprite->dir == 3)
						{
							if (thissprite->tile > 38) thissprite->tile = 36;
						}
						else
						{
							if (thissprite->tile > 41) thissprite->tile = 39;
						}
					}
				}	
				
				// Update robot animation
				else if (thissprite->type == 1)
				{
					thissprite->animCounter++;
					if (thissprite->animCounter > 4)
					{
						thissprite->animCounter = 0;
						
						if (thissprite->dir == 3)
						{						
							if (thissprite->tile == 52) thissprite->tile = 53; else thissprite->tile = 52;
						}
						else // 1
						{
							if (thissprite->tile == 54) thissprite->tile = 55; else thissprite->tile = 54;
						}
					}
				}
								
				
			}

			// Erase sprites from screen
			for (c = 0; c < MAXSPRITES; c++)
			{
				struct sprite_struct *thissprite = &sprites[c];
				if (thissprite->type == -1) continue;
				
				// Erase previous position on screen
				if (bbVisible(thissprite->lastx, thissprite->lasty))
					//blitTileFrom16(thissprite->lastx - screen_xpixeloffset, thissprite->lasty - screen_ypixeloffset, thissprite->background);
				{
					int tilex, tiley, initialx, initialy;
					unsigned char tile;
					
					initialx = tilex = thissprite->lastx >> 4;
					initialy = tiley = thissprite->lasty >> 4;				
					//tileUpdates[tileUpdatePosition++] = tilex | (tiley << 8);
					tile = tileMap[(tiley << 6) + tilex] - 1;  // * MAP_WIDTH
					copyTileFromStorage((tilex - map_startX) << 4, (tiley - map_startY) << 4, tile);					
					
					tilex = (thissprite->lastx + 15) >> 4;
					tiley = thissprite->lasty >> 4;				
					if (initialx != tilex)
					{
						//tileUpdates[tileUpdatePosition++] = tilex | (tiley << 8);
						tile = tileMap[(tiley << 6) + tilex] - 1;
						copyTileFromStorage((tilex - map_startX) << 4, (tiley - map_startY) << 4, tile);
					}
					
					tilex = thissprite->lastx >> 4;
					tiley = (thissprite->lasty + 15) >> 4;				
					if (initialy != tiley)
					{
						//tileUpdates[tileUpdatePosition++] = tilex | (tiley << 8);
						tile = tileMap[(tiley << 6) + tilex] - 1;
						copyTileFromStorage((tilex - map_startX) << 4, (tiley - map_startY) << 4, tile);
					}
 					
					tilex = (thissprite->lastx + 15) >> 4;
					tiley = (thissprite->lasty + 15) >> 4;				
					if (initialx != tilex && initialy != tiley)
					{				
						//tileUpdates[tileUpdatePosition++] = tilex | (tiley << 8);				
						tile = tileMap[(tiley << 6) + tilex] - 1;
						copyTileFromStorage((tilex - map_startX) << 4, (tiley - map_startY) << 4, tile);
					}
				}
			}	
			
			
			// Draw sprites
			for (c = MAXSPRITES-1; c >= 0; c--)
			{			
				struct sprite_struct *thissprite = &sprites[c];
				if (thissprite->type == -1) continue;
				
				if (bbVisible(thissprite->x, thissprite->y))
				{
					//saveTileTo16(thissprite->x - screen_xpixeloffset, thissprite->y - screen_ypixeloffset, thissprite->background);
					blitTile16(thissprite->x - screen_xpixeloffset, thissprite->y - screen_ypixeloffset, thissprite->tile);
					//copyTileFromStorage(thissprite->x - screen_xpixeloffset, thissprite->y - screen_ypixeloffset, thissprite->tile);
				}
			}
		}
		 
		//if (skipRender > 1) skipRender = 0;
		// if (skipRender)
		// {
			// if (frameCounter & 1)
			// {
				// tempPage = visiblePage;
				// visiblePage = nonvisiblePage;
				// nonvisiblePage = tempPage;
			// }
			// skipRender++;
		// }
		// else
		{
			if (frameCounter & 1)
			{
				// Swap page offsets		
				tempPage = visiblePage;
				visiblePage = nonvisiblePage;
				nonvisiblePage = tempPage;
				
				page_addr = visiblePage + (yoffset * scanlineWidthInBytes) + (xoffset >> 2);
				high_addr = HIGH_ADDRESS | (page_addr & 0xff00);
				low_addr = LOW_ADDRESS  | (page_addr << 8);
				
				// wait for previous trace to finish
				//_asm cli
				while (inp(0x3DA) & 0x8) {}	
			
				// switch starting video address
				outpw(CRTC_INDEX, high_addr);
				outpw(CRTC_INDEX, low_addr);	
				
				// set horizontal pixel shift
				inp(0x3DA);
				outp(ACR_INDEX, 0x33); //0x13, 0x33
				outp(ACR_INDEX, (xoffset % 4) << 1);
				
				if (fadeout) writePaletteAll(0, 256, currentPalette);
				else
				{
					if ((markerTimer % 16) > 7) writePaletteSingle(112, 0xFF, 0, 0);
					else writePaletteSingle(112, 0x70, 0, 0);
					markerTimer++;
				}
							
				// wait for next trace to start
				while (!(inp(0x3DA) & 0x8)) {}	
				//_asm sti
			}
			else
			{
				while (inp(0x3DA) & 0x8) {}	
				if (fadeout) writePaletteAll(0, 256, currentPalette);
				else
				{
					if ((markerTimer % 16) > 7) writePaletteSingle(112, 0xFF, 0, 0);
					else writePaletteSingle(112, 0x70, 0, 0);
					markerTimer++;
				}
				while (!(inp(0x3DA) & 0x8)) {}	
			}
			
			if (fadeout == 3) fadeout = 0;	
		}
		
		frameCounter++;
			
		

		
		
		
		//inp(0x3DA);
		//outp(ACR_INDEX, 0x20);
		
    }
	
	//getch();

	_dos_setvect(9, oldkbhandler);
	
    setVideoMode(3);

	printf("Last FPS: %d\n", fps);
	printf("\nDr. Maze\nMini-Ludum Dare 52 entry, June 2014\nBy Jeff Bowman (fyberoptic@gmail.com)\n\nThanks for playing!\n\n");
    
    return 0;
}

